#include "PackageTableSortFilterProxyModel.h"
#include "PackageTable.h"
#include "Package.h"

PackageTable::SortFilterProxyModel::SortFilterProxyModel
   (StateFilter  stateFilter,
    bool         dupsFilter,
    bool         groupsFilter,
    YumIO*       yumIO,
    QObject*     parent)
   : HSortFilterProxyModel(NAME, parent),
     mode(FILTER),
     filter(true),
     stateFilter(stateFilter),
     dupsFilter(dupsFilter),
     groupsFilter(groupsFilter),
     yumIO(yumIO),
     sModel(NULL)
#ifdef USE_REGCOMP
   , regcomp_status(-1)
#endif
{
}

void PackageTable::SortFilterProxyModel::setSourceModel
   (HAbstractTableModel*  sourceModel)
{
#ifdef MODEL_HAS_ABOUTTOBERESET
   // This causes clear() to be called when the sourceModel is aboutToBeReset().
   //
   // Doing this in response to the modelReset() causes it to happen too late;
   // for some reason, Qt queries this model in the midst of a reset(), which
   // could cause it to request and save stale information.
   if (sModel != NULL) {
      disconnect(sModel, SIGNAL(modelAboutToBeReset()),
                 this,   SLOT(clearCache()));
   }
#endif

   HSortFilterProxyModel::setSourceModel(sourceModel);
   sModel = sourceModel;

#ifdef MODEL_HAS_ABOUTTOBERESET
   if (sModel != NULL) {
      connect(sModel, SIGNAL(modelAboutToBeReset()),
              this,   SLOT(clearCache()));
   }
#endif
}

void PackageTable::SortFilterProxyModel::setMode(Mode  mode)
{
   if (this->mode != mode) {
      this->mode = mode;
      in.clear();
      // In Qt 4.2 and higher, it might be possible to just call 
      // filterChanged() here instead.
      clear();
   }
}

void PackageTable::SortFilterProxyModel::setFilter(bool  filter)
{
   if (this->filter != filter) {
      this->filter = filter;
      // In Qt 4.2 and higher, it might be possible to just call 
      // filterChanged() here instead.
      clear();
   }
}

bool PackageTable::SortFilterProxyModel::getFilter ()
{
   return filter;
}

void PackageTable::SortFilterProxyModel::setFilterRegExp
   (const QString&  regExp)
{
   QString  pattern = regExp;
   while (pattern.startsWith(".*")) {
      pattern.remove(0, 2);
   }
   while (pattern.endsWith(".*")) {
      pattern.chop(2);
   }

#ifdef USE_REGCOMP
   if (regcomp_status == 0) {
      regfree(&preg);
   }
   regcomp_status = regcomp(&preg, pattern.toStdString().c_str(), 
                            REG_EXTENDED|REG_NOSUB);
#endif

   QRegExp  re(pattern);
   if (re != filterRegExp()) {
      in.clear();
   }
   HSortFilterProxyModel::setFilterRegExp(re);

   int  row = 0;
   for (; row < sourceModel()->rowCount(); row++) {
      filterAcceptsRow(row, QModelIndex());
   }
}

void PackageTable::SortFilterProxyModel::setStateFilter
   (StateFilter  stateFilter)
{
   if (this->stateFilter != stateFilter) {
      this->stateFilter = stateFilter;
      in.clear();
      // In Qt 4.2 and higher, it might be possible to just call 
      // filterChanged() here instead.
      clear();
   }
}

void PackageTable::SortFilterProxyModel::setDupsFilter(bool  dupsFilter)
{
   if (this->dupsFilter != dupsFilter) {
      this->dupsFilter = dupsFilter;
      in.clear();
      // In Qt 4.2 and higher, it might be possible to just call 
      // filterChanged() here instead.
      clear();
   }
}

void PackageTable::SortFilterProxyModel::setGroupsFilter(bool  groupsFilter)
{
   if (this->groupsFilter != groupsFilter) {
      this->groupsFilter = groupsFilter;
      in.clear();
      // In Qt 4.2 and higher, it might be possible to just call 
      // filterChanged() here instead.
      clear();
   }
}

void PackageTable::SortFilterProxyModel::setUpdateListed ()
{
   int  row = 0;
   for (; row < rowCount(); row++) {
      QModelIndex      proxyIx = index(row, ACTION);
      QModelIndex      srcIx   = mapToSource(proxyIx);
      Package::Entry*  entry   = (*yumIO)[srcIx.row()];
      if (entry->state == Package::OUT_OF_DATE) {
         entry->action = Package::UPDATE;
         emit dataChanged(proxyIx, proxyIx);
      }
   }
}

bool PackageTable::SortFilterProxyModel::filterAcceptsRow
   (int                 source_row, 
    const QModelIndex&  source_parent) const
{
   if (!filter) return true;

   QModelIndex      ix  = sourceModel()->index(source_row, NAME, source_parent);
   Package::Entry*  pkg = sourceModel()->data(ix, RawValue)
                             .value<Package::Entry*>();

   if (in.contains(pkg)) {
      return in[pkg];
   }

   bool  result = acceptRow(pkg, source_row, source_parent);
   in[pkg] = result;
   return result;
}

bool PackageTable::SortFilterProxyModel::matches (const QString&  s) const
{
#ifdef USE_REGCOMP
   return (regcomp_status != 0
           || regexec(&preg, s.toStdString().c_str(), 0, NULL, 0) == 0);
#else
   return s.contains(filterRegExp());
#endif
}

bool PackageTable::SortFilterProxyModel::acceptRow
   (Package::Entry*     pkg,
    int                 source_row,
    const QModelIndex&  source_parent) const
{
   // state filter

   {
      using namespace Package;
      switch (stateFilter) {
      case UPDATABLE:   if (pkg->state != OUT_OF_DATE)   return false; break;
      case INSTALLABLE: if (pkg->state != NOT_INSTALLED) return false; break;
      case REMOVABLE:   if (pkg->state == NOT_INSTALLED) return false; break;
      default:                                                         break;
      }
   }

   // dups filter
   if (!dupsFilter) {
      if (pkg->duplicateName) {
         return false;
      }
   }

   // groups filter

   if (!groupsFilter) {
      if (pkg->kind == Package::GROUP) {
         return false;
      }
   }

   // string filter or search

   switch (mode) {
   case FILTER:
      if (!filterRegExp().isEmpty()) {
         QString  key;

         int  column;
         for (column = NAME; column < NUM_COLUMNS; column++) {
            QModelIndex  ix = sourceModel()->index(source_row, column,
                                                   source_parent);
            if (!key.isEmpty()) key.append(" ");
            key.append(sourceModel()->data(ix, Qt::DisplayRole).toString());
         }

         if (!matches(key)) return false;
      }
      break;
   case SEARCH:
      if (!filterRegExp().isEmpty()) {
         switch (pkg->kind) {
         case Package::SUITE:
         case Package::PACKAGE:
            if (   !matches(yumIO->getName(*pkg))
                && !matches(yumIO->getVerRel(*pkg))
                && !matches(yumIO->getRepo(*pkg))
                && !matches(yumIO->getArch(*pkg))
                && !matches(yumIO->getSize(*pkg))
                && !matches(yumIO->getSummary(*pkg))
                && !matches(yumIO->getDescription(*pkg))
                && !matches(yumIO->getGroup(*pkg))
                && !matches(yumIO->getVendor(*pkg))
                && !matches(yumIO->getBuildTime(*pkg))
                && !matches(yumIO->getBuildHost(*pkg))
                && !matches(yumIO->getInstallTime(*pkg))
                && !matches(yumIO->getBinaryRpm(*pkg))
                && !matches(yumIO->getSourceRpm(*pkg))
                && !matches(yumIO->getLicense(*pkg))
                && !matches(yumIO->getPackager(*pkg))
                && !matches(yumIO->getUrl(*pkg))
                && !matches(yumIO->getRelocations(*pkg))
                && !matches(yumIO->getProvides(*pkg))
                && !matches(yumIO->getRequiresPre(*pkg))
                && !matches(yumIO->getRequires(*pkg))
                && !matches(yumIO->getConflicts(*pkg))
                && !matches(yumIO->getObsoletes(*pkg))
                && !matches(yumIO->getRecommends(*pkg))
                && !matches(yumIO->getSuggests(*pkg))
                && !matches(yumIO->getSupplements(*pkg))
                && !matches(yumIO->getEnhances(*pkg))
                && !matches(yumIO->getFiles(*pkg))
                && !matches(yumIO->getChangeLog(*pkg))) {
               return false;
            }
            break;
         case Package::GROUP:
            if (   !matches(yumIO->getName(*pkg))
                && !matches(yumIO->getMandatoryPkgs(*pkg))
                && !matches(yumIO->getDefaultPkgs(*pkg))
                && !matches(yumIO->getOptionalPkgs(*pkg))
                && !matches(yumIO->getConditionalPkgs(*pkg))) {
               return false;
            }
            break;
         default:
            return false;
         }
      }
      break;
   }

   return true;
}

bool PackageTable::SortFilterProxyModel::lessThan
   (const QModelIndex&  left,
    const QModelIndex&  right) const
{
   switch (left.column()) {
   case ACTION:
   case STATE:
   case KIND:
   case SIZE:
   case BUILDTIME:
      if (left.model() == NULL)  return false;
      if (right.model() == NULL) return false;
      {
         QVariant l = left.model()->data(left, RawValue);;
         QVariant r = right.model()->data(right, RawValue);
         if (l.toInt() < r.toInt()) {
            return true;
         } else if (l.toInt() == r.toInt()) {
            return lessThan(createSecondarySortIndex(left),
                            createSecondarySortIndex(right));
         } else {
            return false;
         }
      }
   default:
      return HSortFilterProxyModel::lessThan(left, right);
   }
}

void PackageTable::SortFilterProxyModel::clear()
{
   in.clear();
   HSortFilterProxyModel::clear();
}

void PackageTable::SortFilterProxyModel::clearCache()
{
   in.clear();
}
